import json
import time
from pathlib import Path
from sys import platform
from typing import List

import soundfile as sf
from click import secho
from noneprompt import ConfirmPrompt
from soundfile import _formats

from .config import config
from .text import Slice, load_text


def lab(audio: Path, slices: List[Slice], des: Path):
    excp = {}
    with sf.SoundFile(audio, "r") as f:
        data = f.read()
        for n, slice in enumerate(slices, start=0):
            # write .wav
            s = int(slice.start * f.samplerate)
            e = int(slice.end * f.samplerate)
            if (
                des.joinpath(f"{audio.stem}_{n:02d}.lab").exists()
                and config.skip_exist_slice
            ):
                secho(f"切片 {audio.stem}_{n:02d} 已存在, 跳过此切片", fg="black")
                continue
            sf.write(
                des.joinpath(f"{audio.stem}_{n:02d}.wav"),
                data[s:e],
                samplerate=f.samplerate,
            )
            # layout lyrics expection
            lyrics, excp_ = slice.get_lyrics(des.joinpath(f"{audio.stem}_{n:02d}.wav"))
            if excp_:
                excp[f'{des.absolute().joinpath(f"{audio.stem}_{n:02d}")}.lab'] = excp_
            # write .lab with lyrics
            des.joinpath(f"{audio.stem}_{n:02d}.lab").write_text(
                lyrics, encoding="utf-8"
            )
    return excp


def main(src: Path, des: Path, subtype: str):
    excp = []
    st = time.time()
    des.mkdir(0o755, parents=True, exist_ok=True)
    filels = sorted(set(src.rglob(f"*.{subtype}")), key=lambda f: f.name, reverse=True)
    for i, ass_file in enumerate(filels):
        secho(
            f"[{i:3}/{len(filels):3}]{int(time.time()-st):5}s: 正在处理 {ass_file}",
            fg="bright_green",
        )
        # check Chinese char
        if (
            platform == "win32"
            and any("\u4e00" <= ch <= "\u9fff" for ch in ass_file.stem)
            and ConfirmPrompt(
                f"{ass_file} 的文件名中包含中文, 强行制作数据集可能会出现错误, 是否跳过该文件?",
                default_choice=True,
            ).prompt()
        ):
            continue
        ass = load_text(ass_file.read_text(encoding="utf-8"), subtype)
        if not ass:
            continue
        for audio_file in ass_file.parent.rglob(f"{ass_file.stem}.*"):
            if audio_file.suffix.upper()[1:] not in _formats.keys():
                continue
            if e := lab(audio_file, ass, des):
                excp.append(e)
            break
    des.joinpath("exception.json").write_text(
        json.dumps(excp, ensure_ascii=False), encoding="utf-8"
    )



import threading
import time
from pathlib import Path

from click import secho
from noneprompt import Choice, ListPrompt,CancelledError

import wave

import pyaudio

from ..config import config
from ..text import Lyrics, Slice
from .tracker import get_freq, log_freq


class LyricsChoice(Choice):
    lyrics: Lyrics

    @classmethod
    def create(cls, lyrics: Lyrics):
        lyrics.pinyin, s = get_freq(lyrics.han, lyrics.pinyin)
        lc = cls(
            f"[{s}]{lyrics.index:>2}: {lyrics.han}({lyrics.pinyin[lyrics._choice]}) => {' '.join(lyrics.pinyin)}"
        )
        lc.lyrics = lyrics
        return lc

    def change_pinyin(self):
        pin = ListPrompt(
            f"您要将 {self.lyrics.han} 的拼音修改为?",
            choices=[Choice(pin) for pin in self.lyrics.pinyin],
            annotation="使用键盘的 ↑ 和 ↓ 来选择, 按回车确认",
        ).prompt()
        self.lyrics.change_by_pinyin(pin.name)
        self.log_freq()

    def log_freq(self):
        log_freq(self.lyrics.han, self.lyrics.pinyin[self.lyrics._choice])


def play_audio(audio: Path):
    try:
        with wave.open(str(audio.absolute()), "rb") as wf:
            p = pyaudio.PyAudio()
            stream = p.open(
                format=p.get_format_from_width(wf.getsampwidth()),
                channels=wf.getnchannels(),
                rate=wf.getframerate(),
                output=True,
            )
            while len(data := wf.readframes(1024)):
                stream.write(data)
            stream.close()
            p.terminate()
    except Exception as e:
        raise RuntimeError("此设备无法正常播放音频") from e


def main_page(slice: Slice, audio: Path):
    secho(f"===== 切片 <{audio.absolute()}> 中包含多音字 =====", fg="bright_red")
    secho(f"原文: {' '.join([ly.han for ly in slice.lyrics_ls])}")
    choices = [
        Choice("[🎵]播放音频"),
    ]
    # append choices
    choices.extend(LyricsChoice.create(ly) for ly in slice.lyrics_ls if ly._warning)
    choices.append(Choice("[🎉]完成编辑"))
    try:
        choice = ListPrompt(
            "请选择您要进行的操作:", choices=choices, annotation="使用 ↑ ↓ 选择, 回车确认 ⚪数据不足 🔴重点校对 🟡优先校对 🟢一般校对"
        ).prompt()
    except CancelledError:
        raise KeyboardInterrupt
    # finish
    if choices[-1] == choice:
        if len(choices) == 2:
            return
        for choice in choices[1:-1]:
            choice: LyricsChoice
            choice.log_freq()
    # play audio
    elif choices[0] == choice:
        threading.Thread(target=play_audio, args=(audio,)).start()
        time.sleep(0.5)
        return main_page(slice, audio)
    # change pinyin
    else:
        choice: LyricsChoice
        choice.change_pinyin()
        return main_page(slice, audio)



import json
from typing import List

from click import secho

from pathlib import Path

from ..config import config


TRACKER_PATH = Path(config.tracker_path)
if not (TRACKER_PATH.exists() and TRACKER_PATH.is_file()):
    if config.tracker_download:
        secho(
            f"{TRACKER_PATH} 不存在或不为文件! 将尝试从 {config.tracker_download} 下载此文件!",
            fg="bright_yellow",
        )
        import urllib.request

        try:
            urllib.request.urlretrieve(config.tracker_download, TRACKER_PATH)
            TRACKER_PATH.chmod(0o755)
            secho("下载成功!", fg="bright_green")
        except Exception as e:
            secho(f"下载失败! 错误原因: {e}", fg="bright_red")
            secho("自动在此位置创建空白文件!", fg="bright_red")
            TRACKER_PATH.touch(0o755, exist_ok=True)
            TRACKER_PATH.write_text("{}", encoding="utf-8")
    else:
        secho(f"{TRACKER_PATH} 不存在或不为文件! 将自动在此位置创建空白文件!", fg="bright_red")
        TRACKER_PATH.touch(0o755, exist_ok=True)
        TRACKER_PATH.write_text("{}", encoding="utf-8")

# tracker 是可以被导入的外置词典, 权重较低
tracker: dict = json.loads(TRACKER_PATH.read_text(encoding="utf-8"))
# tracker_local 是仅内存的内置词典, 权重较高, 更能反映当前用户的选择倾向
tracker_local: dict = {}


def get_freq(word: str, pinyin: List[str]):
    w_ = tracker.get(word) or {}
    w = tracker_local.get(word) or w_

    total = sum(w_.values())
    # 若总数不大于 5 则白色
    if total < 5:
        s = "⚪"
    # 若两词典的最大值不统一则红色
    elif (
        sorted(w_.keys(), key=lambda x: w_[x], reverse=True)[0]
        != sorted(w.keys(), key=lambda x: w[x], reverse=True)[0]
    ):
        s = "🔴"

    # 若最大值选项占比大于 97% 则绿色, 反之黄色
    else:
        s = "🟢" if any(x / total > 0.97 for x in w_.values()) else "🟡"

    try:
        return sorted(pinyin, key=lambda x: w.get(x, 0), reverse=True), s
    except Exception:
        return pinyin, w


def log_freq(word: str, pinyin: str):
    global tracker
    global tracker_local
    for t in (tracker, tracker_local):
        w = t.get(word)
        # if no record
        if not w:
            t[word] = {pinyin: 1}
        # if no pinyin
        elif not w.get(pinyin):
            w[pinyin] = 1
            t[word] = w
        else:
            w[pinyin] += 1
            t[word] = w
    TRACKER_PATH.write_text(json.dumps(tracker, ensure_ascii=False), encoding="utf-8")



from .slice import Slice
from typing import List


def _time_to_sec(t: str):
    h, m, s = t.split(":")
    return int(h) * 3600 + int(m) * 60 + float(s)


def load(ass: str):
    ls: List[Slice] = []
    for l in ass.split("\n"):
        if not l.startswith("Dialogue: "):
            continue
        i = l[10:].strip().split(",", 9)
        s = Slice(
            start=_time_to_sec(i[1]), end=_time_to_sec(i[2]), lyrics_text=i[-1]
        )
        ls.append(s)
    return ls



from typing import Union, List
from ..config import config

from pypinyin import pinyin, Style
from pathlib import Path

from noneprompt import ConfirmPrompt

from click import secho


class Lyrics:
    index: int
    han: str
    pinyin: List[str]
    _choice: int = 0
    _warning: bool = True

    def __init__(self, index: int, han: str, pinyin: List[str]) -> None:
        self.index: int = index
        self.han: str = han
        self.pinyin: List[str] = pinyin
        if len(pinyin) == 1:
            self._warning = False

    def get(self):
        return self.pinyin[self._choice], self._warning

    def change_by_choice(self, choice: int):
        if choice < len(self.pinyin):
            self._choice = choice
            self._warning = False

    def change_by_pinyin(self, pinyin: str):
        try:
            self._choice = self.pinyin.index(pinyin)
            self._warning = False
        except ValueError:
            return


class Slice:
    """Slice 对象包含了时间信息与对应歌词，用于切片及生成输出文件"""

    start: Union[int, float]
    """开始时间"""
    end: Union[int, float]
    """结束时间"""
    dur: Union[int, float]
    """持续时间"""
    lyrics_ls: List[Lyrics]
    """歌词"""

    def __init__(
        self, start: Union[int, float], end: Union[int, float], lyrics_text: str
    ):
        # check slice duration
        dur = end - start
        if dur < config.slice_min_sec or dur > config.slice_max_sec:
            raise ValueError(
                f"切片时长必须在 {config.slice_min_sec} 至 {config.slice_max_sec} 秒之间, 而不是为 {dur:.2f} 秒, 跳过此文件"
            )

        # init param
        self.start = start
        self.end = end
        self.dur = dur
        self.lyrics_ls = []
        i = 0
        is_ascii = False
        for x in lyrics_text:
            if x in [" "]:
                is_ascii = False
            elif x.isascii():
                if not is_ascii:
                    self.lyrics_ls.append(Lyrics(i, x, [x]))
                    i += 1
                    is_ascii = True
                else:
                    ly = self.lyrics_ls.pop()
                    ly.han += x
                    ly.pinyin = [ly.han]
                    self.lyrics_ls.append(ly)
            else:
                is_ascii = False
                self.lyrics_ls.append(
                    Lyrics(
                        i,
                        x,
                        pinyin(
                            x,
                            heteronym=config.pinyin_heteronym_check,
                            style=Style.NORMAL,
                        )[0],
                    )
                )
                i += 1

    def get_lyrics(self, audio: Path):
        if config.pinyin_interactive_check and any(
            lyrics._warning for lyrics in self.lyrics_ls
        ):
            from ..interactive import main_page

            main_page(self, audio)

        lys = []
        excp = []
        for lyrics in self.lyrics_ls:
            pinyin, warning = lyrics.get()
            lys.append(pinyin)
            if warning:
                excp.append(f"{lyrics.han} -> {lyrics.pinyin}")

        return " ".join(lys), excp



from pathlib import Path

import click
from noneprompt import CancelledError

@click.command(help="将音频文件按照字幕文件进行切片, 并将字幕文件转换成 .lab 文件")
@click.option(
    "-i",
    "--input",
    prompt="请输入需要处理的文件夹的路径",
    help="需要处理的文件夹",
)
@click.option(
    "-o",
    "--output",
    prompt="请输入处理后输出的文件夹的路径",
    help="处理后输出的文件夹",
)
@click.option(
    "-s",
    "--subtype",
    prompt_required=False,
    type=click.Choice(["ass"], case_sensitive=False),
    prompt="请输入字幕文件类型",
    help="字幕文件类型",
    default="ass",
)
@click.option("-f", "--force", is_flag=True, help="强制执行")
def slicer(input: str, output: str, subtype: str, force: bool):
    # sourcery skip: raise-from-previous-error
    src = Path(input)
    des = Path(output)
    if not src.exists():
        click.secho(f"源路径 {src.absolute()} 不存在！", fg="bright_red")
        return
    elif not src.is_dir():
        click.secho(f"源路径 {src.absolute()} 不是文件夹！", fg="bright_red")
        return
    elif not len(list(src.iterdir())):
        click.secho(f"源路径 {src.absolute()} 是空文件夹！", fg="bright_red")
        return

    des.mkdir(0o755, True, True)
    if len(list(des.iterdir())) and not force:
        click.secho(f"目标路径 {des.absolute()} 不是空文件夹！请添加 --force 或 -f 跳过此检查")
        return

    try:
        from ..functions.slicer import main as slicer

        slicer(src, des, subtype)
    except CancelledError:
        raise KeyboardInterrupt



import sys

if sys.version_info < (3, 10):
    from importlib_metadata import EntryPoint, version, entry_points
else:
    from importlib.metadata import EntryPoint, version, entry_points

try:
    __version__ = version("nb-cli")
except Exception:
    __version__ = None


import json
import os
from pathlib import Path
from typing import Any

from click import secho
from noneprompt import CancelledError, Choice, ConfirmPrompt, InputPrompt, ListPrompt

DEFULAT_CONFIG_PATH = (
    Path(os.getcwd()).joinpath("voices_config", "slicer.json").absolute()
)


class Config:

    """Slice Setting"""

    slice_min_sec: float = 0
    slice_max_sec: float = 15

    """pinyin setting"""
    pinyin_heteronym_check: bool = True
    pinyin_interactive_check: bool = False

    """tracker setting"""
    tracker_download: str = "https://ali.well404.top/files/tracker.json"
    tracker_path: str = str(DEFULAT_CONFIG_PATH.parent.joinpath("slicer_tracker.json"))

    """skip setting"""
    skip_exist_slice: bool = True

    def __dir__(self):
        return [
            "slice_min_sec",
            "slice_max_sec",
            "pinyin_heteronym_check",
            "pinyin_interactive_check",
            "tracker_download",
            "tracker_path",
            "skip_exist_slice",
        ]

    def __init__(self) -> None:
        if DEFULAT_CONFIG_PATH.exists():
            self.load(defualt=True)

        choices = [
            Choice("[🔧]修改配置"),
            Choice("[📃]导入配置"),
            Choice("[📝]导出配置"),
            Choice("[🎉]完成修改"),
        ]

        while True:
            secho(
                "\n".join(
                    [
                        "===== 当前配置项 =====",
                        f"最小切片时长: {self.slice_min_sec}",
                        f"最大切片时长: {self.slice_max_sec}",
                        f"多音字检查: {self.pinyin_heteronym_check}",
                        f"交互式多音字检查: {self.pinyin_heteronym_check}",
                        f"多音字提示词典路径: {self.tracker_path}",
                        f"多音字提示词典下载: {self.tracker_download}",
                        f"跳过已存在的切片: {self.skip_exist_slice}",
                    ]
                )
            )
            choice = ListPrompt("请导出/载入/调整配置项, 或直接开始:", choices=choices).prompt()
            if choice == choices[0]:
                self.modify()
            elif choice == choices[1]:
                self.load()
            elif choice == choices[2]:
                self.save()

            elif choice == choices[3]:
                break

    def modify(self):
        while True:
            try:
                min_sec = float(
                    f"""{float(
                    InputPrompt(
                        "请输入 **切片最短时长**, 取值范围为 [0, 5] 秒:", default_text=str(self.slice_min_sec)
                    ).prompt()
                ):.02f}"""
                )
                if 0 <= min_sec <= 5:
                    self.slice_min_sec = min_sec
                    break
            except CancelledError:
                secho("已终止配置流程! 已修改的部分将会保留!", fg="bright_red")
                return
            except Exception as e:
                secho(f"输入值无法解析, 请重新输入: {e}", fg="bright_red")

        while True:
            try:
                max_sec = float(
                    f"""{float(
                    InputPrompt(
                        "请输入 **切片最长时长**, 取值范围为 (5, 20] 秒:", default_text=str(self.slice_max_sec)
                    ).prompt()
                ):.02f}"""
                )
                if 5 < max_sec <= 20:
                    self.slice_max_sec = max_sec
                    break
            except CancelledError:
                secho("已终止配置流程! 已修改的部分将会保留!", fg="bright_red")
                return
            except Exception as e:
                secho(f"输入值无法解析, 请重新输入: {e}", fg="bright_red")

        try:
            self.pinyin_heteronym_check = ConfirmPrompt(
                "是否开启多音字检查?", default_choice=self.pinyin_heteronym_check
            ).prompt()
            if self.pinyin_heteronym_check:
                self.pinyin_interactive_check = ConfirmPrompt(
                    "是否开启 **交互式** 多音字检查?", default_choice=self.pinyin_interactive_check
                ).prompt()
            else:
                self.pinyin_interactive_check = False

            self.tracker_download = InputPrompt(
                "请输入 <tracker.json 在线地址>(若不想在线获取则留空)?",
                default_text=self.tracker_download,
            ).prompt()
            self.tracker_path = InputPrompt(
                "请输入 <tracker.json 保存路径>?", default_text=self.tracker_path
            ).prompt()

            self.skip_exist_slice = ConfirmPrompt(
                "是否跳过已存在的切片?", default_choice=self.skip_exist_slice
            ).prompt()
        except CancelledError:
            secho("已终止配置流程! 已修改的部分将会保留!", fg="bright_red")
            return

    def load(self, defualt: bool = False):
        if defualt:
            secho("检测到当前路径包含 slicer 的配置文件, 尝试读取")
            try:
                ans = DEFULAT_CONFIG_PATH.read_text(encoding="utf-8")
            except Exception as e:
                secho(f"无法正常读取配置文件, 使用默认值: {e}")
                return
        else:
            ans = InputPrompt(
                "请输入配置文件的路径, 或直接输入其内容:", default_text=str(DEFULAT_CONFIG_PATH)
            ).prompt()
            if ans.strip().startswith("{"):
                secho("检测到输入为配置信息, 尝试解析")
            else:
                secho("检测到输入为配置文件路径, 尝试读取")
                try:
                    ans = Path(ans).read_text(encoding="utf-8")
                except Exception as e:
                    secho(f"无法正常读取配置文件: {e}")
                    self.load()

        try:
            cfg: dict[str, Any] = json.loads(ans.strip())
            secho("解析成功, 正在覆盖当前配置", fg="bright_green")
        except CancelledError:
            return
        except Exception as e:
            secho(f"无法解析配置信息, 请重新输入: {e}", fg="bright_red")
            self.load()
        for key in cfg:
            self.__setattr__(key, cfg[key])

    def save(self):
        cfg = json.dumps(
            {str(x): self.__getattribute__(x) for x in self.__dir__()},
            ensure_ascii=False,
        )
        try:
            ans = InputPrompt(
                "请输入要保存的路径, 或输入 str 以直接获取其内容", default_text=str(DEFULAT_CONFIG_PATH)
            ).prompt()
            if ans.lower() == "str":
                print(cfg)
                return
            des = Path(ans)
            des.parent.mkdir(0o755, True, True)
            des.write_text(cfg, encoding="utf-8")
            return
        except CancelledError:
            return
        except Exception as e:
            secho(f"无法正常写入, 请重新输入: {e}", fg="bright_red")
            self.save()


config: Config = Config()



from typing import Literal

from click import secho

from . import ass
from .slice import Lyrics, Slice

support_type = Literal["ass"]


def load_text(text: str, text_type: support_type):
    try:
        if text_type == "ass":
            return ass.load(text)
        else:
            raise NotImplementedError
    except Exception as e:
        for _ in range(3):
            secho(e,fg="bright_red")
